#
# Coupled tanks example: Multiple shooting
# Master thesis attachment
# Konstantin Nalum 2013
#

import numpy as NP
from matplotlib.pyplot import *
from pymodelica import compile_jmu,compile_fmu
from pyjmi import JMUModel
from casadi import *
from pyfmi import FMUModel
import zipfile
import scipy.io
import os

os.system("CLS")

# Compile model and generate XML equations
jmu_name = compile_jmu("ocp","tank.mop",'optimica','ipopt',{'generate_xml_equations':True, 'generate_fmi_me_xml':False})
sfile = zipfile.ZipFile('ocp.jmu','r')
mfile = sfile.extract('modelDescription.xml','.')

# Obtain Symbolic OCP representation
ocp = SymbolicOCP()
ocp.parseFMI('modelDescription.xml',{'scale_variables':True,'sort_equations':False})	# Scale system equations
ocp.sortType()

tf=600		# Final time
nk=60		# Number of segments

nz=2+1		# States (model eq + cost state)
nu = 1		# Control inputs

def toArray(v, der=False):	# Convert symbolic OCP variables to another format
  ret = []
  for i in v:
    if der:
      ret.append(i.der())
    else:
      ret.append(i.var())
  return NP.array(ret,dtype=SX)
  
ocp.makeExplicit()

z = toArray(ocp.xd)	# States
u = toArray(ocp.u)	# Control inputs
f=ocp.ode			# ODE RHS (model equations)

# Generate ode: f=(z,u)
ode=SXFunction(daeIn(z,[],u,[],[]),daeOut(f))	


f_d = CVodesIntegrator(ode)						# Instantiate a CVodesIntegrator
f_d.setOption("abstol",1e-5)					# Absolute tolerance
f_d.setOption("reltol",1e-5)					# Relative tolerance
f_d.setOption("steps_per_checkpoint",1000)		# Steps per checkpoint
f_d.setOption("tf",tf/nk)						# Integration length for each u[k]
f_d.init()


nv = nu*nk + nz*(nk)

V = msym("V", nv)								# Decision variable vector [u,z]

U = V[0:nk*nu]
Z0 = V[nk*nu+0:nv:nz]
Z1 = V[nk*nu+1:nv:nz]
Z2 = V[nk*nu+2:nv:nz]

# Variable bounds
VMIN = (-inf)*NP.ones(nv)
VMAX = (inf)*NP.ones(nv)

VMIN[0:nk*nu] 		=	0		# u_min
VMAX[0:nk*nu] 		=	0.3		# u_max

VMIN[nk*nu+1:nv:nz]	=	0.1		# z1_min
VMIN[nk*nu+2:nv:nz]	=	0.1		# z2_min

VMAX[nk*nu+1:nv:nz]	=	5		# z1_max
VMAX[nk*nu+2:nv:nz]	=	5		# z2_max


# Initial conditions
C0	=	0
z10	=	ocp.variable('z1').getStart()
z20	=	ocp.variable('z2').getStart()

# Initial guess vector
#u_guess = NP.random.rand()
u_guess = 0.3
initU	=	NP.ones(nk*nu)*u_guess
initV2	=	NP.transpose(NP.tile([C0,z10,z20],(nk)))
initVec	=	NP.concatenate((initU,initV2))
VINIT	=	initVec

# Constraint functions
g = []
g_min=NP.zeros([nk,nz])
g_max=NP.zeros([nk,nz])

g.append(vertcat((Z0[0],Z1[0],Z2[0]))-vertcat((C0,z10,z20)))		# Initial condition constr.

# Build up a graph of integrator calls
for k in range(nk):
  Zk = vertcat((Z0[k],Z1[k],Z2[k]))
  
  if k!=nk-1:
	Zk_next = vertcat((Z0[k+1],Z1[k+1],Z2[k+1]))
  
  Zk_end = f_d.call([Zk,U[k]])[0]				# Integrator call
  
  if k!= nk-1:
	g.append(Zk_next - Zk_end)					# Continuity constraints


F = MXFunction([V],[Zk_end[0]])					# Zk_end[0] = Objective function: zc(tf)=zc(tN)
G = MXFunction([V],[vertcat(g)])				# Constraint fcn g(V)

solver = IpoptSolver(F,G)						# Provide obj. and constr. to IPOPT
solver.setOption("tol",1e-1)					# IPOPT tolerance
solver.setOption("max_iter",500)				# IPOPT maximum iterations
solver.init()

solver.setInput(VMIN,  NLP_LBX)					# u_L
solver.setInput(VMAX,  NLP_UBX)					# u_U
solver.setInput(VINIT, NLP_X_INIT)				# Initial guess

G_MIN = vertcat((NP.concatenate(g_min)))
G_MAX = vertcat((NP.concatenate(g_max)))

solver.setInput(G_MIN,NLP_LBG)	# g_L
solver.setInput(G_MAX,NLP_UBG)	# g_U

solver.solve()					# Solve NLP


# Obtain results from optimization
v_opt = solver.output(NLP_X_OPT)

u_opt = v_opt[0:nk]


z0_opt = v_opt[nk*nu+0:nv:nz]
z1_opt = v_opt[nk*nu+1:nv:nz]
z2_opt = v_opt[nk*nu+2:nv:nz]


tgrid_z = NP.linspace(0,tf,nk)
tgrid_u = NP.linspace(0,tf,nk)


clf()

subplot(211)
plot(tgrid_z,z0_opt,'r')
plot(tgrid_z,z1_opt,'g')

subplot(212)
plot(tgrid_u,u_opt,'b')

grid()
#show()

u_opt=NP.array(u_opt)
t_opt=NP.linspace(0,tf,nk)

# Compile model for simulation purpose
fmu_name = compile_fmu("tank","tank.mop")
vdp = FMUModel(fmu_name)
t = NP.linspace(0.,tf,nk) 

# Create u profile for inputting to simulation
u_traj = NP.transpose(NP.vstack((t,NP.transpose(u_opt))))
input_object = (['u'], u_traj)

opts = vdp.simulate_options()
opts['solver'] = 'CVode'
opts['ncp'] = nk-1
res = vdp.simulate(final_time=tf, input=input_object,options=opts)

z1=res['z1']
z2=res['z2']


u=res['u']

t=res['time']

# Export data to MATLAB
# fname='C:/'
# scipy.io.savemat(fname,dict(	t_ms=t,z1_ms=z1,z2_ms=z2,u_ms=u,u_ms_opt=u_opt))


# Plot optimized and simulated profiles
clf()
subplot(211)
plot(t,z1,'r')
plot(t,z2,'g')

subplot(212)
step(t,u,'b')
step(t_opt,u_opt,'r')

show()