#
# Coupled tanks example: Single shooting
# Master thesis attachment
# Konstantin Nalum 2013
#

import numpy as NP
from matplotlib.pyplot import *
from pymodelica import compile_jmu,compile_fmu
from pyjmi import JMUModel
from casadi import *
from pyfmi import FMUModel
import zipfile
import scipy.io

import os
os.system("CLS")

# Compile model and generate XML equations
jmu_name= compile_jmu("ocp","tank.mop",'optimica','ipopt',{'generate_xml_equations':True})
sfile 	= zipfile.ZipFile('ocp.jmu','r')
mfile 	= sfile.extract('modelDescription.xml','.')

# Obtain Symbolic OCP representation
ocp = SymbolicOCP()
ocp.parseFMI('modelDescription.xml',{'scale_variables':True,'sort_equations':False})	# Scale system equations

tf=ocp.tf	# Final time
nk=60		# Control input discretization

def toArray(v, der=False):	# Convert symbolic OCP variables to another format
  ret = []
  for i in v:
    if der:
      ret.append(i.der())
    else:
      ret.append(i.var())
  return NP.array(ret,dtype=SX)
  
ocp.makeExplicit()

z	=	toArray(ocp.xd)	# States
u	=	toArray(ocp.u)	# Control inputs
f	=	ocp.ode			# ODE RHS (model equations)


# Generate ode: f=(z,u)
ode=SXFunction(daeIn(z,[],u,[],[]),daeOut(f))	

f_d = CVodesIntegrator(ode)						# Instantiate a CVodesIntegrator
f_d.setOption("abstol",1e-5) 					# Absolute tolerance
f_d.setOption("reltol",1e-5) 					# Relative tolerance
f_d.setOption("steps_per_checkpoint",1000)		# Steps per checkpoint
f_d.setOption("tf",tf/nk) 						# Integration length for each u[k]
f_d.init()

# Initial conditions
C0 	=	0
z10	=	ocp.variable('z1').getStart()
z20	=	ocp.variable('z2').getStart()

U = msym("U",nk)								# All control inputs in a vector
Z = msym([C0,z10,z20])							# Initial states in a vector

Z1 = MX("Z1",nk)								# All z1 states in a vector
Z2 = MX("Z2",nk)								# All z2 states in a vector

# Build up a graph of integrator calls
for k in range(nk):
  [Z,_,_,_] = f_d.call([Z,U[k]])	# Integrator call
  Z1[k] = Z[1]						# State z1
  Z2[k] = Z[2]						# State z2
  
F = MXFunction([U],[Z[0]])				# Z[0] = Objective function: zc(tf)=zc(tN)
G = MXFunction([U],[vertcat((Z1,Z2))])	# Constraint fcn g(Z)

solver = IpoptSolver(F,G)				# Provide obj. and constr. to IPOPT
solver.setOption('tol',0.1)				# IPOPT tolerance
solver.setOption('max_iter',500)		# IPOPT maximum iterations
solver.init()

solver.setInput(0*NP.ones(nk), NLP_LBX)		# u_L
solver.setInput(0.3*NP.ones(nk), NLP_UBX)	# u_U

u_guess = 0.1
#u_guess = NP.random.rand() # random guess
solver.setInput(NP.ones(nk)*u_guess,NLP_X_INIT)	# Initial guess

solver.setInput(vertcat((NP.ones(nk)*0.1,NP.ones(nk)*0.1)),NLP_LBG)	# g_L
solver.setInput(vertcat((NP.ones(nk)*5,NP.ones(nk)*5)),NLP_UBG)		# g_U


solver.solve()	# Solve NLP


# Obtain results from optimization
t_opt = NP.linspace(0,tf,nk)
u_opt = NP.array(solver.output(NLP_X_OPT))

# Compile model for simulation purpose
fmu_name = compile_fmu("tank","tank.mop")
vdp = FMUModel(fmu_name)
t = NP.linspace(0.,tf,nk) 

# Create u profile for inputting to simulation
u_traj = NP.transpose(NP.vstack((t,NP.transpose(u_opt))))
input_object = (['u'], u_traj)

opts = vdp.simulate_options()
opts['solver'] = 'CVode'
opts['ncp'] = nk-1

res = vdp.simulate(final_time=tf, input=input_object,options=opts)

# Plot optimized and simulated profiles
z1=res['z1']
z2=res['z2']
u=res['u']
t=res['time']

clf()
subplot(2,1,1)
plot(t,z1,'b')
plot(t,z2,'g')

subplot(2,1,2)
step(t,u,'b--')
step(t_opt,u_opt,'r-.')

# Export data to MATLAB
# fname='C:/'
# scipy.io.savemat(fname,dict(t_ss=t,z1_ss=z1,z2_ss=z2,u_ss=u,u_ss_opt=u_opt))

show()
